##############################################################################
# Module : utility.py
# Author :
# Date   :
# Updated and modified for ABC_OKC : Old King Cole
#
# Description : ABC_OKC parameters and various utilities
#
##############################################################################
import wx, sys, socket

sysencoding = sys.getfilesystemencoding()

from lang.lang import Lang, ABC_OKC_BUILD
from os import rename, remove, path, walk, system, mkdir, listdir
from distutils.dir_util import copy_tree
from threading import Thread
from time import strftime
from wx.lib.buttons import GenBitmapButton, GenBitmapToggleButton
from win32api import MessageBox
from win32con import VER_PLATFORM_WIN32_NT, VER_PLATFORM_WIN32_WINDOWS
if sys.getwindowsversion()[3] == VER_PLATFORM_WIN32_NT:
    from win32api import GetCurrentProcess, ExitWindowsEx
    from win32security import OpenProcessToken, LookupPrivilegeValue, AdjustTokenPrivileges
    from win32con import TOKEN_ALL_ACCESS, SE_SHUTDOWN_NAME, SE_PRIVILEGE_ENABLED

from filemanager import TorrentConfigFileManager

def getUserDirectories():
    # Returns user and ABC_OKC data directories
    userdir = path.expanduser('~')
    if not path.isdir(userdir):
        if sys.getwindowsversion()[3] == VER_PLATFORM_WIN32_WINDOWS:
            # Windows 9x
            standarduserappdata = path.join(path.expandvars('${WINDIR}'), 'application data')
            if path.exists(standarduserappdata):
                if path.isdir(standarduserappdata):
                    userdir = standarduserappdata
                else:
                    userdir = path.abspath(path.dirname(sys.argv[0]))
            else:
                mkdir(standarduserappdata)
                userdir = standarduserappdata
        else:
            userdir = path.abspath(path.dirname(sys.argv[0]))
    for var in ['${APPDATA}', '${HOME}', '${HOMEPATH}', '${USERPROFILE}']:
        appdatadir = path.expandvars(var)
        if appdatadir != var and path.isdir(appdatadir):
            break
    else:
        appdatadir = userdir
    return userdir, path.join(appdatadir, 'ABC_OKC')

USERPATH, DATAPATH = getUserDirectories()


class Utility:
    def __init__(self, parent, abcpath):
        self.abcpath = abcpath
        self.parent = parent

        # Invalid characters for file names in Windows
        self.invalidwinfilenamechar = ''
        for i in xrange(32):
            self.invalidwinfilenamechar += chr(i)
        self.invalidwinfilenamechar += '"*/:<>?\\|'

        # System support for IPv6 ?
        self.hasipv6 = sys.version_info >= (2, 3) and socket.has_ipv6

        # List to keep formated data from the column of a listctrl being sorted
        self.sortcol = []

        # Torrent advanced details columns
        try:
            fw = wx.SystemSettings_GetFont(wx.SYS_DEFAULT_GUI_FONT).GetPointSize() + 1
        except:
            fw = wx.SystemSettings_GetFont(wx.SYS_SYSTEM_FONT).GetPointSize() + 1
        self.advdetcolinfo = [('advdetoptunch', 'advdetwidthoptunch', wx.LIST_FORMAT_CENTER, fw * 2 + 4),
                              ('advdetip', 'advdetwidthip', wx.LIST_FORMAT_LEFT, fw * 11),
                              ('advdetlocrem', 'advdetwidthlocrem', wx.LIST_FORMAT_CENTER, fw * 3),
                              ('advdetuprate', 'advdetwidthuprate', wx.LIST_FORMAT_RIGHT, fw * 6),
                              ('advdetinterested', 'advdetwidthinterested', wx.LIST_FORMAT_CENTER, fw * 2 + 2),
                              ('advdetchoking', 'advdetwidthchoking', wx.LIST_FORMAT_CENTER, fw * 2 + 2),
                              ('advdetdnrate', 'advdetwidthdnrate', wx.LIST_FORMAT_RIGHT, fw * 6),
                              ('advdetinteresting', 'advdetwidthinteresting', wx.LIST_FORMAT_CENTER, fw * 2 + 2),
                              ('advdetchoked', 'advdetwidthchoked', wx.LIST_FORMAT_CENTER, fw * 2 + 2),
                              ('advdetsnubbed', 'advdetwidthsnubbed', wx.LIST_FORMAT_CENTER, fw * 2 + 2),
                              ('advdetsent', 'advdetwidthsent', wx.LIST_FORMAT_RIGHT, fw * 7),
                              ('advdetreceived', 'advdetwidthreceived', wx.LIST_FORMAT_RIGHT, fw * 7),
                              ('advdetprogress', 'advdetwidthprogress', wx.LIST_FORMAT_RIGHT, fw * 6),
                              ('advdetpeerdnspeed', 'advdetwidthpeerdnspeed', wx.LIST_FORMAT_RIGHT, fw * 6)]

        # RSS panel columns
        self.rsscolinfo = [('rsstitle', 'rsswidthtitle', wx.LIST_FORMAT_LEFT),
                           ('rssurl', 'rsswidthurl', wx.LIST_FORMAT_LEFT),
                           ('rsstime', 'rsswidthtime', wx.LIST_FORMAT_LEFT)]

        # User and data directories
        self.userpath = USERPATH
        self.datapath = DATAPATH

        # Old data directory
        if sys.getwindowsversion()[3] == VER_PLATFORM_WIN32_WINDOWS:
            # Windows 9x
            self.olddatapath = path.join(path.expandvars('${WINDIR}'), 'all users', 'application data', 'ABC_OKC', 'olddata')
        else:
            if sys.getwindowsversion()[0] >= 6:
                # Vista and above
                applicationdata = ''
            else:
                applicationdata = 'application data'
            self.olddatapath = path.join(path.expandvars('${ALLUSERSPROFILE}'), applicationdata, 'ABC_OKC', 'olddata')

        # Init configuration files
        self.initConfigFiles()

        # Read ABC_OKC parameters
        self.readParams()

        # Read language file
        self.lang = Lang(self.parent, self.abcpath, self.abcparams['language_file'])
        self.localize = self.lang.get 

        # IPv6 enabled by user
        self.ipv6enabled = self.hasipv6 and self.abcparams['ipv6'] == '1'

        # Default main port
        self.mainport = 56666

        # DHT port
        try:
            self.dhtport = int(self.abcparams['dhtport'])
        except:
            self.dhtport = 56666
        else:
            if self.dhtport < 0  or self.dhtport > 65535:
                self.dhtport = 56666

        # GUI update rates
        if self.abcparams['tweaksysparams'] == '0':
            self.initDefaultGUIUpdateRate()
        else:
            try:
                self.GUIupdaterate_fast = float(self.abcparams['guiupdateratefast'])
                self.GUIupdaterate_slow = float(self.abcparams['guiupdaterateslow'])
            except:
                self.initDefaultGUIUpdateRate()
            else:
                if self.GUIupdaterate_fast < 0.5:
                    self.GUIupdaterate_fast = 0.5
                if self.GUIupdaterate_slow < 0.5:
                    self.GUIupdaterate_slow = 0.5
        self.GUIupdaterate = self.GUIupdaterate_fast

        # Colours
        rgb = map(int, self.abcparams['colnocon'].split(','))
        self.colnocon = wx.Colour(rgb[0], rgb[1], rgb[2])
        rgb = map(int, self.abcparams['colok'].split(','))
        self.colok = wx.Colour(rgb[0], rgb[1], rgb[2])
        rgb = map(int, self.abcparams['colnocom'].split(','))
        self.colnocom = wx.Colour(rgb[0], rgb[1], rgb[2])
        rgb = map(int, self.abcparams['coldisc'].split(','))
        self.coldisc = wx.Colour(rgb[0], rgb[1], rgb[2])
        rgb = map(int, self.abcparams['colnoinc'].split(','))
        self.colnoinc = wx.Colour(rgb[0], rgb[1], rgb[2])
        rgb = map(int, self.abcparams['colstripes'].split(','))
        self.colstripes = wx.Colour(rgb[0], rgb[1], rgb[2])
        self.colsyswin = wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW)
        self.colnoeng = wx.Colour(0, 0, 0)

        # Seeding time counter behaviour
        # Count
        # 0 : Always count seeding time whatever seeding option is set ;
        # 1 : Only count seeding time when seeding option is time limit ;
        self.seedingtimecount = self.abcparams['seedingtimecount']
        # Reset time count on time limit value change
        # 0 : Don't reset ;
        # 1 : Reset ;
        self.seedingtimeresetonvalue = self.abcparams['seedingtimeresetonvalue']
        # Reset time count on switch to time limit option
        # 0 : Don't reset ;
        # 1 : Reset ;
        self.seedingtimeresetonswitch = self.abcparams['seedingtimeresetonswitch']

        # Scheduler cyclical tasks rate
        if self.abcparams['tweaksysparams'] == '0':
            self.initDefaultCyclicalTasksRate()
        else:
            try:
                self.cyclicaltasksrate = float(self.abcparams['cyclicaltasksrate'])
            except:
                self.initDefaultCyclicalTasksRate()
            else:
                if self.cyclicaltasksrate < 0.5:
                    self.cyclicaltasksrate = 0.5

        # Torrent list auto save rate
        try:
            self.torlistautosaverate = float(self.abcparams['torlistautosaverate'])
        except:
            self.torlistautosaverate = 600.0
        else:
            if self.torlistautosaverate < 60:
                self.torlistautosaverate = 60.0

        # PEX message sending rate
        try:
            self.utpexsendingrate = int(self.abcparams['utpexsendingrate'])
        except:
            self.utpexsendingrate = 90
        else:
            if self.utpexsendingrate < 60:
                self.utpexsendingrate = 60

        # Delay before URM starts after starting of ABC_OKC
        try:
            self.urmstartdelay = int(self.abcparams['urmstartdelay'])
        except:
            self.urmstartdelay = 180

        # Minimum number of uploads
        try:
            self.minuploads = int(self.abcparams['minuploads'])
        except:
            self.minuploads = 2
        else:
            if self.minuploads < 0:
                self.minuploads = 0
            elif self.minuploads > 100:
                self.minuploads = 100

        # Minimum up rate
        try:
            self.minuprate = int(self.abcparams['minuprate'])
        except:
            self.minuprate = 3
        else:
            if self.minuprate < 1:
                self.minuprate = 1

        # Auto scrape rate
        try:
            self.scrapeauto = int(self.abcparams['scrapeauto'])
        except:
            self.scrapeauto = 1200
        else:
            if self.scrapeauto < 60:
                self.scrapeauto = 60

        # Standby minimum scrape period
        try:
            self.sbscrape = int(self.abcparams['sbscrape'])
        except:
            self.sbscrape = 1200
        else:
            if self.sbscrape < 60:
                self.sbscrape = 60

        # Standby scrape period per torrent
        try:
            self.sbscrapeperiod = 3600 / float(self.abcparams['sbscrapeph'])
        except:
            self.sbscrapeperiod = 120.

        # Delay to wait for leechers to connect after a standby uncomplete torrent with leechers was resumed
        try:
            self.sbwaitleechers = int(self.abcparams['sbwaitleechers'])
        except:
            self.sbwaitleechers = 120
        else:
            if self.sbwaitleechers < 30:
                self.sbwaitleechers = 30

        # Torrent file list height in torrent details window
        try:
            self.torrentfilelistheight = int(self.abcparams['torrentfilelistheight'])
        except:
            self.torrentfilelistheight = 106

        # Torrent file list widths in torrent details window
        try:
            self.torrentfilelistwidth0 = int(self.abcparams['torrentfilelistwidth0'])
        except:
            self.torrentfilelistwidth0 = 22
        try:
            self.torrentfilelistwidth3 = int(self.abcparams['torrentfilelistwidth3'])
        except:
            self.torrentfilelistwidth3 = 50
        try:
            self.torrentfilelistwidth4 = int(self.abcparams['torrentfilelistwidth4'])
        except:
            self.torrentfilelistwidth4 = 220
        try:
            self.torrentfilelistwidth5 = int(self.abcparams['torrentfilelistwidth5'])
        except:
            self.torrentfilelistwidth5 = 220

        # Torrent tracker list height in torrent details window
        try:
            self.torrenttrackerlistheight = int(self.abcparams['torrenttrackerlistheight'])
        except:
            self.torrenttrackerlistheight = 60

        # Torrent tracker list widths in torrent details window
        try:
            self.torrenttrackerlistwidth0 = int(self.abcparams['torrenttrackerlistwidth0'])
        except:
            self.torrenttrackerlistwidth0 = 47
        try:
            self.torrenttrackerlistwidth1 = int(self.abcparams['torrenttrackerlistwidth1'])
        except:
            self.torrenttrackerlistwidth1 = 0

        # Torrent find result list widths
        try:
            self.abcfindlistwidth0 = int(self.abcparams['abcfindlistwidth0'])
        except:
            self.abcfindlistwidth0 = 40

        # Unsatisfied upload rate threshold that will trigger the stopping of a URM torrent
        # in the scheduler upload rate distribution PHASE 3, low priority urm algo
        try:
            self.missingupratetostopurm = float(self.abcparams['missingupratetostopurm'])
        except:
            self.missingupratetostopurm = 1.0

        # Algo to choose the extra torrent to be queued/on-hold
        # 0 : Choose the torrent with the lower upload rate influence
        # 1 : Choose the torrent which when stopped will bring the upload rate closer to the max upload rate
        #     from below while being in the upload rate dead band, or closer to the max upload rate from above,
        #     or closer to the max upload rate minus the dead band from below.
        try:
            self.haltextraalgo = int(self.abcparams['haltextraalgo'])
        except:
            self.haltextraalgo = 1

        # Whether or not to average the measured total upload and download rates
        self.meantotalrates = (self.abcparams['meantotalrates'] == "1")

        # Parameters for the formula to estimate the real upload rate from the download rate and the upload rate
        # and the number of connections
        if self.abcparams['tweaksysparams'] == '0':
            self.initDefaultDynMaxUpRate()
        else:
            try:
                self.tweakdynmaxuprate = int(self.abcparams['tweakdynmaxuprate'])
                self.urmupfromdownA = float(self.abcparams['urmupfromdownA'])
                self.urmupfromdownB = float(self.abcparams['urmupfromdownB'])
                self.urmupfromdownC = float(self.abcparams['urmupfromdownC'])
                self.urmupfromdownD = float(self.abcparams['urmupfromdownD'])
                self.urmupfromdownE = float(self.abcparams['urmupfromdownE'])
                self.urmupfromdownF = float(self.abcparams['urmupfromdownF'])
                self.urmupfromdownG = float(self.abcparams['urmupfromdownG'])
            except:
                self.initDefaultDynMaxUpRate()

        # Parameters to tweak the bandwidth distribution
        if self.abcparams['tweaksysparams'] == '0':
            self.initDefaultBandwidthDistribution()
        else:
            try:
                self.tweakbandwidthdistribution = int(self.abcparams['tweakbandwidthdistribution'])
                # Bandwidth between reserved down rate and measured down rate over which a torrent will have its reserved down rate lowered
                self.bdcalcdownth1 = float(self.abcparams['bdcalcdownth1'])
                # Bandwidth between reserved down rate and measured down rate under which a torrent will have its reserved down rate raised
                self.bdcalcdownth2 = float(self.abcparams['bdcalcdownth2'])
                # Bandwidth between reserved up rate and measured up rate over which a torrent will have its reserved up rate lowered
                self.bdcalcupth1 = float(self.abcparams['bdcalcupth1'])
                # Bandwidth between reserved up rate and measured up rate under which a torrent will have its reserved up rate raised
                self.bdcalcupth2 = float(self.abcparams['bdcalcupth2'])
                self.bdcalcdownth3 = float(self.abcparams['bdcalcdownth3'])
                self.bdcalcupth3 = float(self.abcparams['bdcalcupth3'])
                self.bdstep1 = float(self.abcparams['bdstep1'])
                self.bdstep2 = float(self.abcparams['bdstep2'])
                self.bdavth2 = float(self.abcparams['bdavth2'])
                # Minimum step for upload and download rates in the scheduler upload rate distribution PHASE 1
                self.bdminstepuprate = float(self.abcparams['bdminstepuprate'])
                self.bdminstepdownrate = float(self.abcparams['bdminstepdownrate'])
            except:
                self.initDefaultBandwidthDistribution()

        # Garbage collector
        if self.abcparams['tweaksysparams'] == '0':
            self.initDefaultGarbageCollector()
        else:
            try:
                self.gcth1 = int(self.abcparams['gcth1'])
                self.gcth2 = int(self.abcparams['gcth2'])
                self.gcth3 = int(self.abcparams['gcth3'])
                self.cyclicalgcrate = int(self.abcparams['cyclicalgcrate']) * 60
            except:
                self.initDefaultGarbageCollector()
            else:
                if self.cyclicalgcrate and self.cyclicalgcrate < 60:
                    self.cyclicalgcrate = 60

        # Metadata timeout
        if self.abcparams['tweaksysparams'] == '0':
            self.initDefaultMetadataTimeout()
        else:
            try:
                self.metadatatimeout = int(self.abcparams['metadatatimeout'])
            except:
                self.initDefaultMetadataTimeout()
            else:
                if self.metadatatimeout < 5:
                    self.metadatatimeout = 5

        # Read/write buffer expiration
        if self.abcparams['tweaksysparams'] == '0':
            self.initDefaultBufferExpiration()
        else:
            try:
                self.readbufferexpiration = int(self.abcparams['readbufferexpiration'])
                self.readbufferpoolexpiration = int(self.abcparams['readbufferpoolexpiration'])
                self.writebufferexpiration = int(self.abcparams['writebufferexpiration'])
            except:
                self.initDefaultBufferExpiration()

        # Time stamp
        self.timestamp = self.abcparams['timestamp']
        self.timestamplen = len(strftime(self.timestamp))

        # Tracker connection error messages
        self.trackercnxerrors = [self.localize('cantreachtracker'),
                                 self.localize('rejectedtracker'),
                                 self.localize('baddatatracker'),
                                 self.localize('nodatatracker')]

        # Status for file names inside torrent
        self.filenamestatus = [self.localize('filenamecharnotfixed'),
                               self.localize('filenametoolongname'),
                               self.localize('filenametoolongpath'),
                               self.localize('filenametoolongdestpath'),
                               self.localize('filenameduplicatenotfixed')]

        # Units
        self.mbps = self.localize('MBps')
        self.kbps = self.localize('kBps')
        self.b = self.localize('Byte')
        self.kb = self.localize('kB')
        self.mb = self.localize('MB')
        self.gb = self.localize('GB')
        self.tb = self.localize('TB')
        self.d = self.localize('days')

        self.priorities_s = [self.localize('highest_s'),
                             self.localize('high_s'),
                             self.localize('normal_s'),
                             self.localize('low_s'),
                             self.localize('lowest_s')]
        self.partialmarker = self.localize('partialmarker')
        self.activitymarker = self.localize('activitymarker')
        self.inactivitymarker = self.localize('inactivitymarker')
        self.standbymarker = self.localize('standbymarker')
        self.notrackermarker = self.localize('notrackermarker')
        self.notrackeranswermarker = self.localize('notrackeranswermarker')
        self.scrapingmarker = self.localize('scrapingmarker')

        # To speed up lists construction :
        self.advdetcolumns = []
        for colinfo in self.advdetcolinfo:
            self.advdetcolumns.append([self.localize(colinfo[0]), int(self.abcparams[colinfo[1]]), colinfo[2]])
        self.rsscolumns = []
        for colinfo in self.rsscolinfo:
            self.rsscolumns.append([self.localize(colinfo[0]), int(self.abcparams[colinfo[1]]), colinfo[2]])

    def readParams(self):
        self.abcparams = {}

        # Read ABC config file
        tcfm = TorrentConfigFileManager(path.join(self.datapath, "abc.conf"), self.abcparams)
        tcfm.readAllConfig()

        abcparamsdefaults = [
            ('scrape', '0'),
            ('defaultpriority', '2'),
            ('language_file', 'english.lang'),
            ('urm', '0'),
            ('urmmaxtorrent', '8'),
            ('urmthresholdstart', '3'),
            ('urmthresholdstop', '1'),
            ('urmlowpriority', '0'),
            ('urmonhold', '1'),
            ('urmstopsman', '0'),
            ('urmdelay', '4'),
            ('urmdelaystop', '5'),
            ('urmdynmaxuprate', '1'),
            ('urmmaxsysuprate', '95'),
            ('urmupfromdownA', '0.027'),
            ('urmupfromdownB', '0.16'),
            ('urmupfromdownC', '0.08'),
            ('urmupfromdownD', '0.04'),
            ('urmupfromdownE', '0.15'),
            ('urmupfromdownF', '0.1'),
            ('urmupfromdownG', '140.8'),
            ('urmstartdelay', '180'),
            ('urmtorrentstartdelay', '60'),
            ('urmtorrentstopdelay', '30'),
            ('stripedlist', '1'),
            ('mode', '1'),
            ('prioritizelocal', '1'),
            ('prioritizelocaldown', '1'),
            ('askduplicatedest', '0'),
            ('fastresume', '1'),
            ('fastresumefirst', '0'),
            ('keepeta', '1'),
            ('quickmetainfo', '1'),
            ('displayunitsize', '1'),
            ('displayunitrate', '1'),
            ('torrentwhencompleted', '0'),
            ('torrentwhenfinished', '0'),
            ('centerquickdetails', '0'),
            ('mainwindowx', '0'),
            ('mainwindowy', '0'),
            ('showschedcounters', '1'),
            ('templatedefault', 'default'),
            ('lastopentemptab', '0'),
            ('lastopentortab', '0'),
            ('templatedlgx', '-1'),
            ('templatedlgy', '-1'),
            ('conlimiterdlgx', '-1'),
            ('conlimiterdlgy', '-1'),
            ('torrentparamsdlgx', '-1'),
            ('torrentparamsdlgy', '-1'),
            ('abcoptiondlgx', '-1'),
            ('abcoptiondlgy', '-1'),
            ('abcfinddlgx', '-1'),
            ('abcfinddlgy', '-1'),
            ('abctweakdlgx', '-1'),
            ('abctweakdlgy', '-1'),
            ('createtorrentdlgx', '-1'),
            ('createtorrentdlgy', '-1'),
            ('webservicedlgx', '-1'),
            ('webservicedlgy', '-1'),
            ('aboutmedlgx', '-1'),
            ('aboutmedlgy', '-1'),
            ('checklatestdlgx', '-1'),
            ('checklatestdlgy', '-1'),
            ('abcrentordlgx', '-1'),
            ('abcrentordlgy', '-1'),
            ('abcsetdestdlgx', '-1'),
            ('abcsetdestdlgy', '-1'),
            ('schedulerdlgx', '-1'),
            ('schedulerdlgy', '-1'),
            ('scannerdlgx', '-1'),
            ('scannerdlgy', '-1'),
            ('rssscanfreqdlgx', '-1'),
            ('rssscanfreqdlgy', '-1'),
            ('scrolllistud', '0'),
            ('scrolllisttb', '0'),
            ('findtorrentnamebtn', '0'),
            ('findtorrentfilenamebtn', '0'),
            ('findlabelbtn', '0'),
            ('findtrackernamebtn', '0'),
            ('maxdownloadrate', '0'),
            ('activitythdown', '0.3'),
            ('activitythup', '0.3'),
            ('activitydelay', '80'),
            ('schedonhold', '0'),
            ('schedstopsman', '0'),
            ('lastopenabcoptab', '0'),
            ('activitymax', '0'),
            ('diskfullth', '10'),
            ('diskfullman', '1'),
            ('savesizesonexit', '0'),
            ('countonlyalive', '0'),
            ('countonhold', '1'),
            ('urmstoponpause', '0'),
            ('bdcalcdownth1', '0.5'),
            ('bdcalcdownth2', '0.1'),
            ('bdcalcdownth3', '0.1'),
            ('bdcalcupth1', '0.5'),
            ('bdcalcupth2', '0.1'),
            ('bdcalcupth3', '0.1'),
            ('bdstep1', '2.5'),
            ('bdstep2', '0.2'),
            ('bdavth2', '0.0'),
            ('bdminstepuprate', '0'),
            ('bdminstepdownrate', '0'),
            ('tweakdynmaxuprate', '0'),
            ('msgindetails', '0'),
            ('defrentorwithdest', '0'),
            ('lastdestloc', ''),
            ('useragentid', '0'),
            ('useragent1', ''),
            ('useragent2', ''),
            ('useragent3', ''),
            ('userid1', ''),
            ('userid2', ''),
            ('userid3', ''),
            ('savetorrent', '2'),
            ('savetordeffolder', ''),
            ('listfont', ''),
            ('listfontsize', ''),
            ('listfontstyle', ''),
            ('listfontweight', ''),
            ('sorttoronclose', '1'),
            ('maketorforeach', '0'),
            ('movemidbtn', '2'),
            ('stoptrafficbtn', '1'),
            ('stopresallbtn', '0'),
            ('schedulerbtn', '1'),
            ('scannerbtn', '1'),
            ('webservbtn', '1'),
            ('statusbarshift', '4'),
            ('sortbytime', '0'),
            ('upnp', '0'),
            ('mousemode', '1'),
            ('rubmessages', '1'),
            ('rubmsgbutton', '0'),
            ('stoprejected', '1'),
            ('sortinglock', '3'),
            ('erasetrackal', '1'),
            ('dontcheckseedactivity', '1'),
            ('keeptermevtmsg', '1'),
            ('starthome', '0'),
            ('homepage', ''),
            ('rssscanfreq', '10'),
            ('rsssash', '0'),
            ('downvolcheck', '0'),
            ('upvolcheck', '0'),
            ('downvolmax', '0'),
            ('upvolmax', '0'),
            ('downvolcountdur', '1'),
            ('upvolcountdur', '1'),
            ('downvolchecktime', '0'),
            ('upvolchecktime', '0'),
            ('downvol', '0'),
            ('upvol', '0'),
            ('downvolcheckcycle', '-1'),
            ('upvolcheckcycle', '-1'),
            ('displaydownvol', '0'),
            ('displayupvol', '0'),
            ('rejectexceptions', ''),
            ('colnocon', '250,0,0'),
            ('colok', '0,170,0'),
            ('colnocom', '0,0,240'),
            ('coldisc', '140,140,140'),
            ('colnoinc', '160,160,0'),
            ('colstripes', '245,245,245'),
            ('colcust01', '0,170,0'),
            ('colcust02', '0,0,240'),
            ('colcust03', '140,140,140'),
            ('colcust04', '255,255,255'),
            ('colcust05', '255,255,255'),
            ('colcust06', '255,255,255'),
            ('colcust07', '255,255,255'),
            ('colcust08', '255,255,255'),
            ('colcust09', '160,160,0'),
            ('colcust10', '250,0,0'),
            ('colcust11', '255,255,255'),
            ('colcust12', '255,255,255'),
            ('colcust13', '255,255,255'),
            ('colcust14', '255,255,255'),
            ('colcust15', '255,255,255'),
            ('colcust16', '255,255,255'),
            ('flattopbtn', '1'),
            ('flatbotbtn', '1'),
            ('rssautograb', '0'),
            ('rsswidthtitle', '400'),
            ('rsswidthurl', '500'),
            ('rsswidthtime', '400'),
            ('initlang', '0'),
            ('scrollbelowfinished', '0'),
            ('guiupdateratefast', '1'),
            ('guiupdaterateslow', '2'),
            ('cyclicaltasksrate', '2'),
            ('timestamp', '%d-%H:%M'),
            ('canceltimeouts', '0'),
            ('canceltimeoutsmsg', ''),
            ('torlistautosaverate', '600'),
            ('crypto', '1'),
            ('showcheckversion', '1'),
            ('expire_cache_data', '10'),
            ('displayscrapingmsg', '0'),
            ('displayscrapestatus', '0'),
            ('displaycheckstatus', '1'),
            ('sizeinmb', '1'),
            ('lockedcompletedcheck', '0'),
            ('lockedcompletedmax', '5'),
            ('gcth1', '500'),
            ('gcth2', '4'),
            ('gcth3', '4'),
            ('tweaksysparams', '0'),
            ('useridmask1', ''),
            ('useridmask2', ''),
            ('useridmask3', ''),
            ('completedresumedelay', '2'),
            ('seedingtimecount', '1'),
            ('seedingtimeresetonvalue', '0'),
            ('seedingtimeresetonswitch', '0'),
            ('minuploads', '2'),
            ('minuprate', '3'),
            ('showactivitymax', '0'),
            ('activityextra', '1'),
            ('extratobestopped', '2'),
            ('findshortlabelbtn', '0'),
            ('scrapeauto', '1200'),
            ('sbscrape', '1200'),
            ('sbscrapeph', '30'),
            ('sbwaitleechers', '120'),
            ('sbtimeout', '1'),
            ('sbtimeoutvalue', '7'),
            ('sbtimeoutaction', '0'),
            ('compshutseedingtime', '1'),
            ('sbqueued', '0'),
            ('filenamescheckresdlgx', '-1'),
            ('filenamescheckresdlgy', '-1'),
            ('centerdetoneachother', '1'),
            ('displaytimeout', '1'),
            ('scheddonthalt', '1'),
            ('scheddonthaltuprategap', '15'),
            ('torrentfilelistheight', '106'),
            ('torrentfilelistwidth0', '22'),
            ('torrentfilelistwidth3', '50'),
            ('torrentfilelistwidth4', '220'),
            ('torrentfilelistwidth5', '220'),
            ('torrenttrackerlistwidth0', '47'),
            ('torrenttrackerlistwidth1', '0'),
            ('abcfindlistwidth0', '40'),
            ('missingupratetostopurm', '1.0'),
            ('urmstopsonlyactive', '1'),
            ('urmstopsonlylowprio', '0'),
            ('haltextraalgo', '1'),
            ('meantotalrates', '0'),
            ('tweakbandwidthdistribution', '0'),
            ('timeoutwhenextra', '1'),
            ('dht', '0'),
            ('dhtnodeid', ''),
            ('dhtrerequestinterval', '300'),
            ('rerequestinterval', '300'),
            ('torrenttrackerlistheight', '60'),
            ('dhtaddnodedlgx', '-1'),
            ('dhtaddnodedlgy', '-1'),
            ('addpeerdlgx', '-1'),
            ('addpeerdlgy', '-1'),
            ('dhtport', '56666'),
            ('cyclicalgcrate', '5'),
            ('multiparamdefault', '1'),
            ('ctrlscraperescanstrackers', '1'),
            ('keepmagnetname', '0'),
            ('metadatatimeout', '60'),
            ('utpex', '1'),
            ('utpexsendingrate', '90'),
            ('writebuffersize', '2'),
            ('readbuffermaxglobalsize', '64'),
            ('readbufferexpiration', '300'),
            ('readbufferpoolexpiration', '300'),
            ('writebufferexpiration', '300'),
            ('readbuffermemorysaving', '0'),
            ('multihomeipv4', ''),
            ('multihomeipv6', '')
            ]

        for parameter, value in abcparamsdefaults:
            if not self.abcparams.has_key(parameter):
                self.abcparams[parameter] = value

        # Init torrent advanced details window columns width
        for colinfo in self.advdetcolinfo:
            if not self.abcparams.has_key(colinfo[1]):
                self.abcparams[colinfo[1]] = str(colinfo[3])

        # Old format for allocation type
        oldalloctype = ['normal', 'background', 'pre-allocate', 'sparse']
        if not self.abcparams['alloc_type'].isdigit():
            self.abcparams['alloc_type'] = str(oldalloctype.index(self.abcparams['alloc_type']))

        # Old format for move folder option
        if not self.abcparams.has_key('movedatastatus'):
            if self.abcparams.has_key('movefoldercompleted') \
               and self.abcparams.has_key('movefolderfinished'):
                if self.abcparams['movefoldercompleted'] == "2":
                    self.abcparams['movedataonstatus'] = "1"
                    self.abcparams['movedatastatus'] = "0"
                elif self.abcparams['movefolderfinished'] == "2":
                    self.abcparams['movedataonstatus'] = "1"
                    self.abcparams['movedatastatus'] = "1"
                else:
                    self.abcparams['movedataonstatus'] = "0"
                    self.abcparams['movedatastatus'] = "1"
                if self.abcparams['movefoldercompleted'] == "1" or self.abcparams['movefolderfinished'] == "1":
                    self.abcparams['movedataondelete'] = "1"
                else:
                    self.abcparams['movedataondelete'] = "0"
                del self.abcparams['movefoldercompleted']
                del self.abcparams['movefolderfinished']
            else:
                self.abcparams['movedataonstatus'] = "0"
                self.abcparams['movedatastatus'] = "1"
                self.abcparams['movedataondelete'] = "0"
        else:
            if self.abcparams.has_key('movefoldercompleted'):
                del self.abcparams['movefoldercompleted']
            if self.abcparams.has_key('movefolderfinished'):
                del self.abcparams['movefolderfinished']

        # Old format for URM deadband
        if self.abcparams.has_key('urmupthreshold'):
            self.abcparams['urmthresholdstart'] = self.abcparams['urmupthreshold']
            del self.abcparams['urmupthreshold']

        # Old unused keys
        if self.abcparams.has_key('bdavth1'):
            del self.abcparams['bdavth1']
        if self.abcparams.has_key('bdgreedyth'):
            del self.abcparams['bdgreedyth']
        #if self.abcparams.has_key('bdstep2'):
        #    del self.abcparams['bdstep2']
        if self.abcparams.has_key('dhtaddpeerdlgx'):
            del self.abcparams['dhtaddpeerdlgx']
        if self.abcparams.has_key('dhtaddpeerdlgy'):
            del self.abcparams['dhtaddpeerdlgy']

        # setextdefaultfolder setting for using default download folder from external source
        if not self.abcparams.has_key('setextdefaultfolder'):
            self.abcparams['setextdefaultfolder'] = self.abcparams['setdefaultfolder']

        # Web service parameters
        self.webparams = {}

        # Read web service config file
        tcfm = TorrentConfigFileManager(path.join(self.datapath, "webservice.conf"), self.webparams)
        tcfm.readAllConfig()

        webparamsdefaults = [
            ('allow_onhold', '1'),
            ('allow_getparam', '1'),
            ('allow_setparam', '1'),
            ('webcoder', 'utf_8'),
            ('allow_move', '1'),
            ('allow_gettorparam', '1'),
            ('allow_settorparam', '1'),
            ('allow_superseed', '1'),
            ('allow_scheduler', '1')
            ]

        for parameter, value in webparamsdefaults:
            if not self.webparams.has_key(parameter):
                self.webparams[parameter] = value

        # Old format for web service IP
        if self.webparams['webIP'] == "Automatic" or self.webparams['webIP'] == "Automatics":
            self.webparams['webIP'] = '0'
        elif self.webparams['webIP'] == "Loop Back":
            self.webparams['webIP'] = '1'

        # Command scheduler parameters
        self.schedparams = {}

        # Read command scheduler config file
        tcfm = TorrentConfigFileManager(path.join(self.datapath, "scheduler.conf"), self.schedparams)
        tcfm.readAllConfig()

        schedparamsdefaults = [
            ('schedrunpast', '0'),
            ('schedautostart', '0')
            ]

        for parameter, value in schedparamsdefaults:
            if not self.schedparams.has_key(parameter):
                self.schedparams[parameter] = value

        # Scanner parameters
        self.scanparams = {}

        # Read scanner config file
        tcfm = TorrentConfigFileManager(path.join(self.datapath, "scanner.conf"), self.scanparams)
        tcfm.readAllConfig()

        scanparamsdefaults = [
            ('scandir', ''),
            ('scanfreqh', '1'),
            ('scanfreqm', '0'),
            ('scanstartpos', '1'),
            ('scanprio', '2'),
            ('scanstatus', '1'),
            ('scanmovetor', '0'),
            ('scanautostart', '0')
            ]

        for parameter, value in scanparamsdefaults:
            if not self.scanparams.has_key(parameter):
                self.scanparams[parameter] = value

    def initDefaultDynMaxUpRate(self):
        self.tweakdynmaxuprate = 0
        self.urmupfromdownA = 0.027
        self.urmupfromdownB = 0.16
        self.urmupfromdownC = 0.08
        self.urmupfromdownD = 0.04
        self.urmupfromdownE = 0.15
        self.urmupfromdownF = 0.1
        self.urmupfromdownG = 140.8

    def initDefaultBandwidthDistribution(self):
        self.tweakbandwidthdistribution = 0
        self.bdcalcdownth1 = 0.5
        self.bdcalcdownth2 = 0.1
        self.bdcalcdownth3 = 0.1
        self.bdcalcupth1 = 0.5
        self.bdcalcupth2 = 0.1
        self.bdcalcupth3 = 0.1
        self.bdstep1 = 2.5
        self.bdstep2 = 0.2
        self.bdavth2 = 0.0
        self.bdminstepuprate = 0.0
        self.bdminstepdownrate = 0.0

    def initDefaultGarbageCollector(self):
        self.gcth1 = 500
        self.gcth2 = 4
        self.gcth3 = 4
        self.cyclicalgcrate = 300

    def initDefaultCyclicalTasksRate(self):
        self.cyclicaltasksrate = 2.0

    def initDefaultGUIUpdateRate(self):
        self.GUIupdaterate_fast = 1.0
        self.GUIupdaterate_slow = 2.0

    def initDefaultMetadataTimeout(self):
        self.metadatatimeout = 60

    def initDefaultBufferExpiration(self):
        self.readbufferexpiration = 300
        self.readbufferpoolexpiration = 300
        self.writebufferexpiration = 300

    def eta_value(self, n, longformat = True, seconds = True, maxvalue = True):
        if n == -1:
            return '?'
        if maxvalue and n >= 8640000:
            return '>100' + self.lang.get('l_day')

        if longformat:
            day, r1 = divmod(n, 86400)
            hour, r2 = divmod(r1, 3600)
            if day > 0:
                return ('%02d' % day) + self.localize('l_day') + ':' + ('%02d' % hour) + self.localize('l_hour')
            min, sec = divmod(r2, 60)
            if hour > 0 or not seconds:
                return ('%02d' % hour) + self.localize('l_hour') + ':' + ('%02d' % min) + self.localize('l_minute')
            return ('%02d' % min) + self.localize('l_minute') + ':' + ('%02d' % sec) + self.localize('l_second')
        hour, r2 = divmod(n, 3600)
        min, sec = divmod(r2, 60)
        return ('%02d' % hour) + self.localize('l_hour') + ':' + ('%02d' % min) + self.localize('l_minute')

    def onWriteFilteredText(self, event):
        # Filter out the characters "|"  and '\x01' from inputs because these are used as separators in the ABC conf files
        keycode = event.GetKeyCode()
        if keycode != 124 and keycode != 1:
            event.Skip()
            return True
        return False

    def onWriteFilteredText2(self, event):
        # Filter out the characters '\x01' and '\x02' from inputs
        keycode = event.GetKeyCode()
        if keycode != 1 and keycode != 2:
            event.Skip()
            return True
        return False

    def onWriteDigits(self, event):
        # Filter out not digit chars
        keycode = event.GetKeyCode()
        if 47 < keycode < 58 or keycode == wx.WXK_DELETE or keycode == wx.WXK_LEFT \
           or keycode == wx.WXK_RIGHT or keycode == wx.WXK_HOME or keycode == wx.WXK_END \
           or keycode == wx.WXK_BACK or keycode == wx.WXK_INSERT or keycode == 3 or keycode == 22:
            event.Skip()
            return True
        return False

    def onWriteDigitsDot(self, event):
        # Filter out not digit or dot chars
        keycode = event.GetKeyCode()
        if keycode == 46 or 47 < keycode < 58 or keycode == wx.WXK_DELETE or keycode == wx.WXK_LEFT \
           or keycode == wx.WXK_RIGHT or keycode == wx.WXK_HOME or keycode == wx.WXK_END \
           or keycode == wx.WXK_BACK or keycode == wx.WXK_INSERT or keycode == 3 or keycode == 22:
            event.Skip()
            return True
        return False

    def onWriteDigitsDotMinus(self, event):
        # Filter out not digit or dot or minus chars
        keycode = event.GetKeyCode()
        if keycode == 45 or keycode == 46 or 47 < keycode < 58 or keycode == wx.WXK_DELETE or keycode == wx.WXK_LEFT \
           or keycode == wx.WXK_RIGHT or keycode == wx.WXK_HOME or keycode == wx.WXK_END \
           or keycode == wx.WXK_BACK or keycode == wx.WXK_INSERT or keycode == 3 or keycode == 22:
            event.Skip()
            return True
        return False

    def onWritePrintable(self, event):
        # Filter out chars > 128 and not printable chars
        keycode = event.GetKeyCode()
        if 31 < keycode < 127 or keycode == wx.WXK_DELETE or keycode == wx.WXK_LEFT \
           or keycode == wx.WXK_RIGHT or keycode == wx.WXK_HOME or keycode == wx.WXK_END \
           or keycode == wx.WXK_BACK or keycode == wx.WXK_INSERT or keycode == 3 or keycode == 22:
            event.Skip()
            return True
        return False

    def getPathSize(self, pth):
        if path.isdir(pth):
            size = 0
            for root, dirs, files in walk(pth):
                size += sum(path.getsize(path.join(root, name)) for name in files)
        else:
            size = path.getsize(pth)
        return size

    def sortList(self, item1, item2):
        # Generic function to sort listctrl
        if self.sortcol[item1] < self.sortcol[item2]:
            return -1
        if self.sortcol[item1] > self.sortcol[item2]:
            return 1
        return 0

    def sortListReverse(self, item1, item2):
        # Generic function to sort listctrl in reverse order
        if self.sortcol[item1] < self.sortcol[item2]:
            return 1
        if self.sortcol[item1] > self.sortcol[item2]:
            return -1
        return 0

    def fixWindowsName(self, str, unit = False):
        # Check if str is a valid Windows file name (or unit name if unit is true)
        # If not, returns a fixed name
        strlen = len(str)
        if unit and (strlen != 2 or str[1] != ':'):
            return 'c:'
        if not str or str == '.' or str == '..':
            return '_'
        if unit:
            str = str[0]
            strlen = 1
        if strlen > 255:
            str = str[:255]
            fixed = True
            strlen = 255
        else:
            fixed = False
        fixedname = ''
        i = 0
        while i < strlen:
            c = str[i]
            if c in self.invalidwinfilenamechar:
                fixedname += '_'
                fixed = True
                i += 1
            elif c == ' ' or c == '.':
                spaceordot = c
                i += 1
                while i < strlen:
                    if str[i] != ' ' and str[i] != '.':
                        break
                    spaceordot += c
                    i += 1
                if i == strlen:
                    # Name ends with spaces and dots
                    fixedname += len(spaceordot) * '_'
                    fixed = True
                else:
                    fixedname += spaceordot
            else:
                fixedname += c
                i += 1
        if fixed:
            return fixedname
        return ''

    def checkWinPath(self, parent, pathtocheck):
        if pathtocheck and pathtocheck[-1] == '\\' and pathtocheck != '\\\\':
            pathitems = pathtocheck[:-1].split('\\')
        else:
            pathitems = pathtocheck.split('\\')
        nexttotest = 1
        if self.isPathRelative(pathtocheck):
            # Relative path
            if len(path.join(self.datapath, pathtocheck)) > 259:
                if parent is not None:
                    dlg = wx.MessageDialog(parent, '"' + pathtocheck + '"\n' + self.localize('errortoolongwinpath'),
                                           self.localize('abcokcerror'), wx.ICON_ERROR)
                    dlg.ShowModal()
                    dlg.Destroy()
                return False
            # Empty relative path is allowed
            if pathtocheck == '':
                return True
            fixedname = self.fixWindowsName(pathitems[0])
            if fixedname:
                if parent is not None:
                    dlg = wx.MessageDialog(parent, '"' + pathitems[0] + '"\n' + self.localize('errorinvalidwinname') + fixedname,
                                           self.localize('abcokcerror'), wx.ICON_ERROR)
                    dlg.ShowModal()
                    dlg.Destroy()
                return False
        else:
            # Absolute path
            if len(pathtocheck) > 259:
                if parent is not None:
                    dlg = wx.MessageDialog(parent, '"' + pathtocheck + '"\n' + self.localize('errortoolongwinpath'),
                                           self.localize('abcokcerror'), wx.ICON_ERROR)
                    dlg.ShowModal()
                    dlg.Destroy()
                return False
            # An absolute path must have at least one '\'
            if not '\\' in pathtocheck:
                if parent is not None:
                    dlg = wx.MessageDialog(parent, '"' + pathitems[0] + '"\n' + self.localize('errorinvalidpath'),
                                           self.localize('abcokcerror'), wx.ICON_ERROR)
                    dlg.ShowModal()
                    dlg.Destroy()
                return False
            if pathtocheck[:2] != '\\\\':
                # Not a network path
                fixedname = self.fixWindowsName(pathitems[0], unit = True)
                if fixedname:
                    if parent is not None:
                        dlg = wx.MessageDialog(parent, '"' + pathitems[0] + '"\n' + self.localize('errorinvalidwinunitname') + fixedname,
                                               self.localize('abcokcerror'), wx.ICON_ERROR)
                        dlg.ShowModal()
                        dlg.Destroy()
                    return False
            else:
                # Network path
                nexttotest = 2

        for name in pathitems[nexttotest:]:
            fixedname = self.fixWindowsName(name)
            if fixedname:
                if parent is not None:
                    dlg = wx.MessageDialog(parent, '"' + name + '"\n' + self.localize('errorinvalidwinname') + fixedname,
                                           self.localize('abcokcerror'), wx.ICON_ERROR)
                    dlg.ShowModal()
                    dlg.Destroy()
                return False

        return True

    def frameWindow(self, win, xl = 0, xr = 0, xt = 0, xb = 0, border = 15, midline = 0):
        parent = win.GetParent()
        winpos = win.GetPositionTuple()
        winsize = win.GetSizeTuple()
        xframe = winpos[0] - xl - 3
        yframe = winpos[1] - xt - 3
        wframe = winsize[0] + xl + xr + 6
        hframe = winsize[1] + xt + xb + 6
        if border & 1:
            wx.StaticLine(parent, -1, (xframe, yframe), (-1, hframe), wx.LI_VERTICAL)
        if border & 2:
            wx.StaticLine(parent, -1, (xframe + 1, yframe), (wframe - 1, -1), wx.LI_HORIZONTAL)
        if border & 4:
            wx.StaticLine(parent, -1, (xframe, yframe + hframe - 2), (wframe, -1), wx.LI_HORIZONTAL)
        if border & 8:
            wx.StaticLine(parent, -1, (xframe + wframe - 2, yframe), (-1, hframe), wx.LI_VERTICAL)
        if midline & 1:
            wx.StaticLine(parent, -1, (xframe + 1, yframe + hframe / 2 - 1), (xl, -1), wx.LI_HORIZONTAL)
        if midline & 2:
            wx.StaticLine(parent, -1, (xframe + wframe / 2, yframe - xb), (-1, xb), wx.LI_VERTICAL)
        if midline & 4:
            wx.StaticLine(parent, -1, (xframe + wframe / 2, yframe + hframe - 2), (-1, xt), wx.LI_VERTICAL)
        if midline & 8:
            wx.StaticLine(parent, -1, (xframe + wframe - 2 - xr, yframe + hframe / 2 - 1), (xr, -1), wx.LI_HORIZONTAL)

    def isPathRelative(self, pth):
        if len(pth) < 2 or pth[1] != ':' and pth[:2] != '\\\\':
            return True
        return False

    def completePath(self, pth):
        if not pth:
            return self.datapath
        if self.isPathRelative(pth):
            return path.join(self.datapath, pth)
        return pth

    def findUniqueFileName(self, dest, src = ''):
        # If file dest exists, returns a new name with '_' and 3 digits added
        # If it can't be renamed this way, asks user for a new name
        # Returns None if a new name could not be set
        # If src is 'web' or 'scan' or 'rss', user is not asked if automatic renaming fails
        if not path.exists(dest):
            return dest

        canbeautomaticallyrenamed = True
        if len(dest) > 255:
            if len(path.split(dest)[0]) > 250:
                canbeautomaticallyrenamed = False
            else:
                dest = dest[:255]

        if canbeautomaticallyrenamed:
            # Renaming will add '_' + 3 digit to the original name
            destlen = len(dest)
            i = 1
            # To be sure the dot is inside the file name (and not in the rest of the path)
            if path.split(dest)[1].find('.') == -1:
                dot = destlen
            else:
                dot = dest.rfind('.')
            while i < 1000:
                newdest = dest[:dot] + ('_%03u' % i) + dest[dot:]
                if not path.exists(newdest):
                    break
                i += 1
            if i != 1000:
                return newdest

        if src == 'web' or src == 'scan' or src == 'rss' or src == 'qcl':
            return None

        while True:
            # Impossible to find an automatic new name ; ask user for one
            dl = wx.FileDialog(self.window, self.localize('chooseuniquetorfilename'),\
                               path.join(self.datapath, 'torrent'), name, '*.*', wx.SAVE)
            result = dl.ShowModal()
            if result != wx.ID_OK:
                dl.Destroy()
                return None
            newdest = dl.GetPath()
            dl.Destroy()
            if not path.exists(newdest):
                return newdest

    def makeBitmap(self, bitmap, trans_color = wx.Colour(200, 200, 200)):
        button_bmp = wx.Bitmap(path.join(self.abcpath, 'icons', bitmap), wx.BITMAP_TYPE_BMP)
        button_mask = wx.Mask(button_bmp, trans_color)
        button_bmp.SetMask(button_mask)
        return button_bmp

    def makeBitmapButton(self, panel, bitmap, tooltip, event, trans_color = wx.Colour(200, 200, 200),
                         toggle = False, bitmapselected = None, bitmapselectedflat = None,
                         bitmapdetected = None, bitmapdetectedflat = None, flat = False):
        button_bmp = wx.Bitmap(path.join(self.abcpath, 'icons', bitmap), wx.BITMAP_TYPE_BMP)
        button_bmp.SetMask(wx.Mask(button_bmp, trans_color))
        if bitmap == 'modemanual.bmp':
            self.modebtnmanual = button_bmp
        if bitmapselected:
            buttonselected_bmp = wx.Bitmap(path.join(self.abcpath, 'icons', bitmapselected), wx.BITMAP_TYPE_BMP)
            buttonselected_bmp.SetMask(wx.Mask(buttonselected_bmp, trans_color))
        if bitmapselectedflat:
            buttonselectedflat_bmp = wx.Bitmap(path.join(self.abcpath, 'icons', bitmapselectedflat), wx.BITMAP_TYPE_BMP)
            buttonselectedflat_bmp.SetMask(wx.Mask(buttonselectedflat_bmp, trans_color))
        if bitmapdetected:
            buttondetected_bmp = wx.Bitmap(path.join(self.abcpath, 'icons', bitmapdetected), wx.BITMAP_TYPE_BMP)
            buttondetected_bmp.SetMask(wx.Mask(buttondetected_bmp, trans_color))
        if bitmapdetectedflat:
            buttondetectedflat_bmp = wx.Bitmap(path.join(self.abcpath, 'icons', bitmapdetectedflat), wx.BITMAP_TYPE_BMP)
            buttondetectedflat_bmp.SetMask(wx.Mask(buttondetectedflat_bmp, trans_color))
        ID_BUTTON = wx.NewId()
        btnsize = wx.Size(button_bmp.GetWidth() + 12, button_bmp.GetHeight() + 4)
        if toggle:
            button_btn = GenBitmapToggleButtonDual(panel, ID_BUTTON, button_bmp, size = btnsize, style = wx.SIMPLE_BORDER)
            if bitmapselected:
                button_btn.setBitmapSelectedNotFlat(buttonselected_bmp)
            else:
                button_btn.setBitmapSelectedNotFlat(button_bmp)
            if bitmapselectedflat:
                button_btn.setBitmapSelectedFlat(buttonselectedflat_bmp)
            else:
                button_btn.setBitmapSelectedFlat(button_bmp)
            if bitmapdetected:
                button_btn.setBitmapDetectedNotFlat(buttondetected_bmp)
            else:
                button_btn.setBitmapDetectedNotFlat(button_bmp)
            if bitmapdetectedflat:
                button_btn.setBitmapDetectedFlat(buttondetectedflat_bmp)
            else:
                button_btn.setBitmapDetectedFlat(button_bmp)
        else:
            button_btn = GenBitmapButtonDual(panel, ID_BUTTON, button_bmp, size = btnsize, style = wx.SIMPLE_BORDER)
            if bitmapselected:
                button_btn.SetBitmapSelected(buttonselected_bmp)
            else:
                button_btn.SetBitmapSelected(button_bmp)
        button_btn.SetUseFocusIndicator(False)
        button_btn.setFlat(flat)

        button_btn.SetToolTipString(tooltip)
        panel.Bind(wx.EVT_BUTTON, event, button_btn)
        return button_btn

    def shutDown(self):
        winver = sys.getwindowsversion()[3]
        if winver == VER_PLATFORM_WIN32_NT:
            shutdownthread = Thread(target = self.shutDown2k)
            shutdownthread.daemon = True
            shutdownthread.start()
        elif winver == VER_PLATFORM_WIN32_WINDOWS:
            system('rundll32 shell32.dll,SHExitWindowsEx 9')

    def shutDown2k(self):
            AdjustTokenPrivileges(OpenProcessToken(GetCurrentProcess(), TOKEN_ALL_ACCESS), False,
                                  [(LookupPrivilegeValue(None, SE_SHUTDOWN_NAME), SE_PRIVILEGE_ENABLED)])
            ExitWindowsEx(9)

    def addToolbarIcon(self, parent, toolbar, event, short, long, bitmap1, bitmap2 = None, type = 'simple', firsttime = False):
        bmp1 = self.makeBitmap(bitmap1)
        if bitmap2 is not None:
            bmp2 = self.makeBitmap(bitmap2)
        if firsttime:
            # Find size of images so it will be dynamics
            toolbar.SetToolBitmapSize(wx.Size(bmp1.GetWidth(), bmp1.GetHeight()))
        newid = wx.NewId()
        parent.Bind(wx.EVT_TOOL, event, id = newid)
        if type == 'simple':
            toolbar.AddSimpleTool(newid, bmp1, short, long)
        elif type == 'toggle':
            toolbar.AddCheckTool(newid, bmp1, bmp2, short, long)
        elif type == 'radio':
            toolbar.AddRadioTool(newid, bmp1, wx.NullBitmap, short, long)
        return newid

    def countPathItems(self, filepath):
        # Count the number of items in a path (not ending with a path separator)
        item = filepath
        nbitem = 0
        while item:
            item = path.split(item)[0]
            nbitem += 1
        return nbitem

    def sizeFormat(self, s):
        if s < 1024:
            return '%i ' % s + self.b
        if s < 1048576:
            return '%.2f ' % (s / 1024.) + self.kb
        if s < 1073741824L:
            return '%.2f ' % (s / 1048576.) + self.mb
        if s < 1099511627776L:
            return '%.2f ' % (s / 1073741824.) + self.gb
        return '%.2f ' % (s / 1099511627776.) + self.tb

    def formatedSize(self, size):
        if self.abcparams['sizeinmb'] == '1':
            if self.abcparams['displayunitsize'] == '1':
                return ('%.2f ' % (size / 1048576.)) + self.mb
            else:
                return ('%.2f' % (size / 1048576.))
        else:
            return self.sizeFormat(size)

    def commaFormat(self, s):
        # Format the string s of an integer by grouping digits 3 by 3 separated by a space from the right to the left
        r = str(s)
        for i in xrange(len(r) - 3, 0, -3):
            r = r[:i] + ' ' + r[i:]
        return r

    def formatedRate(self, rate):
        if self.abcparams['displayunitrate'] == '1':
            return ('%.1f ' % rate) + self.kbps
        else:
            return ('%.1f' % rate)

    def encodeWebServiceData(self, data):
        if self.webparams['webcoder'] == 'utf_8':
            return data.encode('utf_8')
        encoded = ''
        for c in data:
            try:
                encoded += c.encode(sysencoding)
            except:
                encoded += chr(ord(c) / 256) + chr(ord(c) % 256)
        return encoded

    def getListClickedColumn(self, list, x):
        # Returns the clicked column number of the list control from the x coordinate of a point inside the list
        currentwidth = -list.GetScrollPos(wx.HORIZONTAL)
        for col in xrange(list.GetColumnCount()):
            currentwidth += list.GetColumnWidth(col)
            if x < currentwidth:
                return col
        return -1

    def getPosCenteredOnWin(self, sizetobecentered, window, aligntop = False):
        # Returns position for a size to be centered on a window, in the extent of the screen display
        x, y, w, h = window.GetRect()
        wtobecentered, htobecentered = sizetobecentered
        wmax, hmax = wx.DisplaySize()
        xcentered = x + int((w - wtobecentered) / 2.)
        if aligntop:
            ycentered = y
        else:
            ycentered = y + int((h - htobecentered) / 2.)
        if xcentered < 0:
            xcentered = 0
        elif xcentered > wmax - wtobecentered:
            xcentered = wmax - wtobecentered
        if ycentered < 0:
            ycentered = 0
        elif ycentered > hmax - htobecentered:
            ycentered = hmax - htobecentered
        return (xcentered, ycentered)

    def initConfigFiles(self):
        if path.exists(path.join(self.datapath, 'abc.conf')):
            # User conf already exists, just delete old log files
            abcokcbuild = int(ABC_OKC_BUILD)
            logfiles = [f for f in listdir(self.datapath) if f.endswith('.log') and path.isfile(path.join(self.datapath, f))]
            for logfile in logfiles:
                try:
                    logfilebuild = int(logfile.split('.')[-2])
                except:
                    logfilebuild = -1
                if logfilebuild < abcokcbuild:
                    try:
                        remove(path.join(self.datapath, logfile))
                    except:
                        pass
        else:
            # Copy old data or default config files to data path
            if path.exists(self.olddatapath):
                copy_tree(self.olddatapath, self.datapath)
            else:
                copy_tree(path.join(self.abcpath, 'defaultconf'), self.datapath)
                torrentpath = path.join(self.datapath, 'torrent')
                if not path.exists(torrentpath):
                    mkdir(torrentpath)
            # Copy old cache data if any to data path
            appdatadir = path.dirname(self.datapath)
            datacache = path.join(appdatadir, '.ABC', 'datacache')
            if path.exists(datacache):
                copy_tree(datacache, path.join(self.datapath, 'datacache'))
            else:
                mkdir(path.join(self.datapath, 'datacache'))
            piececache = path.join(appdatadir, '.ABC', 'piececache')
            if path.exists(piececache):
                copy_tree(piececache, path.join(self.datapath, 'piececache'))
            else:
                mkdir(path.join(self.datapath, 'piececache'))

    def expandIPv6(self, ip):
        if ip == '::':
            ip = ''
        elif ip[:2] == '::':
            ip = ip[1:]
        elif ip[0] == ':':
            return None
        elif ip[-2:] == '::':
            ip = ip[:-1]
        elif ip[-1] == ':':
            return None
        blocs = ip.split(':')
        dc = None
        blocslen = 0
        i = 0
        for b in blocs:
            if b == '':
                # Double colon
                if dc is not None:
                    return None
                dc = i
            elif '.' in b:
                # IPv4 format
                blocs4 = b.split('.')
                if len(blocs4) != 4:
                    return None
                try:
                    blocs[i] = ':'.join([''.join(['%02x' % int(b4) for b4 in blocs4[j:j + 2]]) for j in xrange(0, 4, 2)])
                except:
                    return None
                blocslen += 2
            elif len(b) < 4:
                blocs[i] = '0' * (4 - len(b)) + b
                blocslen += 1
            else:
                blocslen += 1
            i += 1
        if dc is not None:
            blocs[dc] = ':'.join(['0000'] * (8 - blocslen))
        return ':'.join(blocs)

    def getHostByName(self, name):
        if self.ipv6enabled:
            try:
                host = socket.getaddrinfo(name, None, socket.AF_INET6)[0][4][0]
            except:
                pass
            else:
                return host
        try:
            host = socket.getaddrinfo(name, None, socket.AF_INET)[0][4][0]
        except:
            return None
        return host

    def getIpsByName(self, name):
        try:
            ai = socket.getaddrinfo(name, None)
        except:
            return []
        return [info[4][0] for info in ai]


class GenBitmapToggleButtonDual(GenBitmapToggleButton):
    def setBitmapSelectedFlat(self, bitmap):
        self.bitmapselectedflat = bitmap

    def setBitmapSelectedNotFlat(self, bitmap):
        self.bitmapselectednotflat = bitmap

    def setBitmapDetectedFlat(self, bitmap):
        self.bitmapdetectedflat = bitmap

    def setBitmapDetectedNotFlat(self, bitmap):
        self.bitmapdetectednotflat = bitmap

    def setFlat(self, flat = True):
        if flat:
            self.SetBezelWidth(0)
            self.bitmapselected = self.bitmapselectedflat
            self.bitmapdetected = self.bitmapdetectedflat
        else:
            self.SetBezelWidth(2)
            self.bitmapselected = self.bitmapselectednotflat
            self.bitmapdetected = self.bitmapdetectednotflat
        self.SetBitmapSelected(self.bitmapselected)


class GenBitmapButtonDual(GenBitmapButton):
    def setFlat(self, flat = True):
        if flat:
            self.SetBezelWidth(0)
        else:
            self.SetBezelWidth(2)


class FileTree:
    # To store temporarily the torrent file/folder structure
    def __init__(self, filesname, dest, utility, checkcase = False):
        self.utility = utility
        self.dest = dest
        self.tree = ({}, [])
        self.checkcase = checkcase
        self.nbfiles = len(filesname)
        i = 0
        for file in filesname:
            if self.checkcase:
                items = file.split('\\')
            else:
                items = file.lower().split('\\')
            subitems = self.tree
            for dir in items[:-1]:
                if not dir in subitems[0]:
                    subitems[0][dir] = ({}, [])
                subitems = subitems[0][dir]
            subitems[1].append((i, items[-1]))
            i += 1

    def subItems(self, dirpath):
        if dirpath:
            items = dirpath.split('\\')
        else:
            items = []
        subitems = self.tree
        for dir in items:
            subitems = subitems[0].get(dir)
            if subitems is None:
                return ({}, [])
        return subitems

    def files(self, dirpath):
        return [filetuple[1] for filetuple in self.subItems(dirpath)[1]]

    def dirs(self, dirpath):
        return self.subItems(dirpath)[0]

    def hasFile(self, filepath):
        if self.checkcase:
            dir, file = path.split(filepath)
        else:
            dir, file = path.split(filepath.lower())
        return (file in self.files(dir))

    def hasDir(self, dirpath):
        if self.checkcase:
            updir, dir = path.split(dirpath)
        else:
            updir, dir = path.split(dirpath.lower())
        return (dir in self.dirs(updir))

    def findUniqueWinName(self, name, namelist, pth):
        if name not in namelist and not path.exists(path.join(pth, name)):
            return name
        if len(name) > 251:
            # Unable to rename to make unique without truncating
            return None

        namelen = len(name)
        # Renaming will add '_' + 3 digit to the original name
        i = 1
        dot = name.rfind('.')
        if dot == -1:
            dot = namelen
        while i < 1000:
            newname = name[:dot] + ('_%03u' % i) + name[dot:]
            if newname not in namelist and not path.exists(path.join(pth, newname)):
                break
            i += 1
        if i != 1000:
            return newname
        return None

    def digForFixingWinNames(self, dir, status, pth):
        # Updates status[X, [Y1, Y2,...]] :
        # X : True if at least one name was changed in the tree
        #     False if all names are already OK and none were changed
        # Yi : bit 0 : set if automatically finding unique file name failed while fixing bad char in file i
        #      bit 1 : set if file name is too long in file i
        #      bit 4 : set if automatically finding unique file name failed while fixing duplicate in file i
        # Check directory names
        for d in dir[0].keys():
            if len(d) > 255:
                indexes = []
                self.digForFileIndexes(dir[0][d], indexes)
                for i in indexes:
                    status[1][i] |= 2
            else:
                fixedname = self.utility.fixWindowsName(d)
                if fixedname:
                    status[0] = True
                    uniquefixedname = self.findUniqueWinName(fixedname, dir[0].keys() + [file[1] for file in dir[1]], pth)
                    if uniquefixedname:
                        dir[0][uniquefixedname] = dir[0][d]
                        del dir[0][d]
                        d = uniquefixedname
                    else:
                        indexes = []
                        self.digForFileIndexes(dir[0][d], indexes)
                        for i in indexes:
                            status[1][i] |= 2
            self.digForFixingWinNames(dir[0][d], status, path.join(pth, d))
        # Check file names
        for f in dir[1][:]:
            if len(f[1]) > 255:
                status[1][f[0]] |= 2
            else:
                fixedname = self.utility.fixWindowsName(f[1])
                if fixedname:
                    status[0] = True
                    uniquefixedname = self.findUniqueWinName(fixedname, [file[1] for file in dir[1]] + dir[0].keys(), pth)
                    if uniquefixedname:
                        dir[1].append((f[0], uniquefixedname))
                        dir[1].remove(f)
                    else:
                        status[1][f[0]] |= 1
        # Check duplicate file names
        for i in xrange(len(dir[1])):
            for j in xrange(i):
                if dir[1][j][1] == dir[1][i][1]:
                    status[0] = True
                    uniquefixedname = self.findUniqueWinName(dir[1][i][1], [file[1] for file in dir[1]] + dir[0].keys(), pth)
                    if uniquefixedname:
                        dir[1][i] = (dir[1][i][0], uniquefixedname)
                    else:
                        status[1][dir[1][i][0]] |= 16
                        status[1][dir[1][j][0]] |= 16

    def fixWinNames(self):
        # Check validity of Windows names, and rename if necessary
        status = [False, self.nbfiles * [0]]
        self.digForFixingWinNames(self.tree, status, self.utility.completePath(self.dest))
        return status

    def digForFileIndexes(self, dir, indexes):
        for d in dir[0]:
            self.digForFileIndexes(dir[0][d], indexes)
        indexes.extend([f[0] for f in dir[1]])

    def digForFiles(self, dir, files, curpath = ''):
        for d in dir[0]:
            self.digForFiles(dir[0][d], files, path.join(curpath, d))
        files.extend([(f[0], path.join(curpath, f[1])) for f in dir[1]])
        curpath = path.split(curpath)[0]

    def getFileNames(self):
        # Rebuild the file list from the tree and returns it
        allfiletuples = []
        self.digForFiles(self.tree, allfiletuples)
        allfiletuples.sort()
        return [f[1] for f in allfiletuples]


class SpinBox():
    def __init__(self, parent, header = None, spacing = None, width = 25, value = "0", min = 0, max = 99, filterfunc = None, tooltip = None):
        self.min = min
        self.max = max
        if not filterfunc:
            filterfunc = self.onWriteDigits
        if spacing is None:
            if header:
                spacing = 5
            else:
                spacing = 0
        self.box = wx.TextCtrl(parent, -1, value, (-1, -1), (width, -1), wx.TE_RIGHT)
        self.box.Bind(wx.EVT_CHAR, filterfunc)
        h = self.box.GetSize().height
        self.spin = wx.SpinButton(parent, -1, (-1, -1), (h * 2 / 3, h), wx.SP_ARROW_KEYS | wx.SP_VERTICAL)
        self.spin.SetRange(min, max)
        self.spin.SetValue(int(value))
        self.spin.Bind(wx.EVT_SPIN_UP, self.onSpinUp)
        self.spin.Bind(wx.EVT_SPIN_DOWN, self.onSpinDown)
        if tooltip:
            self.box.setToolTipString(tooltip)
        self.sizer = wx.BoxSizer(wx.HORIZONTAL)
        if header:
            self.sizer.Add(wx.StaticText(parent, -1, header), 0, wx.ALIGN_CENTER_VERTICAL)
        self.sizer.Add(self.box, 0, wx.ALIGN_CENTER_VERTICAL | wx.LEFT, spacing)
        self.sizer.Add(self.spin, 0, wx.ALIGN_CENTER_VERTICAL)

    def onSpinUp(self, event):
        try:
            val = int(self.box.GetValue())
        except:
            # Not integer
            val = self.min
        else:
            if val == self.max:
                return
            if val > self.max:
                val = self.max
            elif val < self.min:
                val = self.min
            else:
                val += 1
        self.spin.SetValue(val)
        self.box.SetValue(str(val))

    def onSpinDown(self, event):
        try:
            val = int(self.box.GetValue())
        except:
            # Not integer
            val = self.min
        else:
            if val == self.min:
                return
            if val < self.min:
                val = self.min
            elif val > self.max:
                val = self.max
            else:
                val -= 1
        self.spin.SetValue(val)
        self.box.SetValue(str(val))

    def onWriteDigits(self, event):
        # Filter out not digit chars
        keycode = event.GetKeyCode()
        if 47 < keycode < 58 or keycode == wx.WXK_DELETE or keycode == wx.WXK_LEFT \
           or keycode == wx.WXK_RIGHT or keycode == wx.WXK_HOME or keycode == wx.WXK_END \
           or keycode == wx.WXK_BACK or keycode == wx.WXK_INSERT or keycode == 3 or keycode == 22:
            event.Skip()
            return True
        return False

    def setToolTipString(self, tooltip):
        self.box.SetToolTipString(tooltip)

    def setValue(self, value):
        try:
            val = int(value)
        except:
            # Not integer
            val = self.min
        else:
            if value < self.min:
                val = self.min
            if val > self.max:
                val = self.max
        self.spin.SetValue(val)
        self.box.SetValue(str(val))

    def getValue(self):
        return self.box.GetValue()

    def getSizer(self):
        return self.sizer

    def getTextWindow(self):
        return self.box

    def getSpinWindow(self):
        return self.spin


# Redirect py2exe error logging to Windows user ABC_OKC app data directory
class Stderr(object):
    softspace = 0
    _file = None
    _error = None
    def write(self, text, alert = MessageBox, fname = sys.executable + '.' + ABC_OKC_BUILD + '.log'):
        if self._file is None and self._error is None:
            if not path.exists(DATAPATH):
                mkdir(DATAPATH)
            fname = path.join(DATAPATH, path.basename(fname))
            try:
                self._file = open(fname, 'a')
            except Exception, details:
                self._error = details
                import atexit
                atexit.register(alert, 0,
                                "The logfile '%s' could not be opened:\n %s" % \
                                (fname, details),
                                "Errors occurred")
            else:
                import atexit
                atexit.register(alert, 0,
                                "See the logfile '%s' for details" % fname,
                                "Errors occurred")
        if self._file is not None:
            self._file.write(text.encode('utf_8'))
            self._file.flush()

    def flush(self):
        if self._file is not None:
            self._file.flush()
